# Optuna tutorial code.
#
# The code is adapted from PyTorch's "Training a Classifier" tutorial:
# https://pytorch.org/tutorials/beginner/blitz/cifar10_tutorial.html
import os
import pathlib
import typing

import optuna
import torch
import torch.nn.functional as F
import torchvision
import torchvision.transforms.v2 as T
from optuna.trial import TrialState
from torch import nn
from torch.nn.modules.utils import consume_prefix_in_state_dict_if_present
from torch.nn.parallel import DistributedDataParallel as DDP  # noqa: N817

import helios.core as hlc
import helios.data as hld
import helios.model as hlm
import helios.nn as hln
import helios.optim as hlo
import helios.trainer as hlt
from helios.core import logging
from helios.plugins.optuna import OptunaPlugin


class CIFARDataModule(hld.DataModule):
    def __init__(self, root: pathlib.Path, batch_size: int):
        self._root = root / "data"
        self._batch_size = batch_size

    def prepare_data(self):
        torchvision.datasets.CIFAR10(root=self._root, train=True, download=True)
        torchvision.datasets.CIFAR10(root=self._root, train=False, download=True)

    def setup(self) -> None:
        transforms = T.Compose(
            [
                hld.transforms.ToImageTensor(),
                T.Normalize((0.5, 0.5, 0.5), (0.5, 0.5, 0.5)),
            ]
        )

        params = hld.DataLoaderParams()
        params.batch_size = self._batch_size
        params.shuffle = True
        params.num_workers = 2
        params.drop_last = True
        self._train_dataset = self._create_dataset(
            torchvision.datasets.CIFAR10(
                root=self._root,
                train=True,
                download=False,
                transform=transforms,
            ),
            params,
        )

        params.drop_last = False
        params.shuffle = False
        params.is_distributed = False
        self._valid_dataset = self._create_dataset(
            torchvision.datasets.CIFAR10(
                root=self._root,
                train=False,
                download=False,
                transform=transforms,
            ),
            params,
        )


@hln.NETWORK_REGISTRY.register
class Net(nn.Module):
    def __init__(self, l1: int = 120, l2: int = 84):
        super().__init__()

        self._conv1 = nn.Conv2d(3, 6, 5)
        self._pool = nn.MaxPool2d(2, 2)
        self._conv2 = nn.Conv2d(6, 16, 5)
        self._fc1 = nn.Linear(16 * 5 * 5, l1)
        self._fc2 = nn.Linear(l1, l2)
        self._fc3 = nn.Linear(l2, 10)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self._pool(F.relu(self._conv1(x)))
        x = self._pool(F.relu(self._conv2(x)))
        x = torch.flatten(x, 1)
        x = F.relu(self._fc1(x))
        x = F.relu(self._fc2(x))
        x = self._fc3(x)
        return x


class ClassifierModel(hlm.Model):
    def __init__(self, batch_size: int):
        super().__init__("classifier")
        self._tune_params: dict[str, float | int] = {}
        self._tune_params["batch_size"] = batch_size

    @property
    def metrics(self) -> dict[str, typing.Any]:
        if not self._val_scores or self._val_scores["total"] == 0:
            return {}

        total = self._val_scores["total"]
        correct = self._val_scores["correct"]
        accuracy = 100 * correct // total
        return {"accuracy": accuracy}

    def setup(self, fast_init: bool = False) -> None:
        plugin = self.trainer.plugins[OptunaPlugin.plugin_id]
        assert isinstance(plugin, OptunaPlugin)

        # Assign the tunable parameters so we can log them as hyper-parameters when
        # training ends.
        self._tune_params["l1"] = plugin.trial.suggest_categorical(
            "l1", [2**i for i in range(9)]
        )
        self._tune_params["l2"] = plugin.trial.suggest_categorical(
            "l2", [2**i for i in range(9)]
        )
        self._tune_params["lr"] = plugin.trial.suggest_float("lr", 1e-4, 1e-1, log=True)

        self._net = Net(
            l1=self._tune_params["l1"],  # type: ignore[arg-type]
            l2=self._tune_params["l2"],  # type: ignore[arg-type]
        ).to(self.device)

        if self.is_distributed:
            self._net = DDP(
                self._net,
                device_ids=[self.rank],
                output_device=self.rank,
            )  # type: ignore[assignment]

        self._criterion = nn.CrossEntropyLoss().to(self.device)
        self._optimizer = hlo.create_optimizer(
            "SGD",
            self._net.parameters(),
            lr=self._tune_params["lr"],
            momentum=0.9,
        )

    def load_state_dict(
        self, state_dict: dict[str, typing.Any], fast_init: bool = False
    ) -> None:
        if not self.is_distributed:
            consume_prefix_in_state_dict_if_present(state_dict["net"], "module.")

        self._net.load_state_dict(state_dict["net"])
        self._optimizer.load_state_dict(state_dict["optimizer"])

    def state_dict(self) -> dict[str, typing.Any]:
        return {"net": self._net.state_dict(), "optimizer": self._optimizer.state_dict()}

    def train(self) -> None:
        self._net.train()

    def train_step(self, batch: typing.Any, state: hlt.TrainingState) -> None:
        inputs: torch.Tensor
        labels: torch.Tensor

        inputs, labels = batch
        inputs = inputs.to(self.device)
        labels = labels.to(self.device)

        self._optimizer.zero_grad()

        outputs = self._net(inputs)
        loss = self._criterion(outputs, labels)
        loss.backward()
        self._optimizer.step()  # type: ignore[operator]

        self._loss_items["loss"] = loss

    def on_training_batch_end(
        self, state: hlt.TrainingState, should_log: bool = False
    ) -> None:
        super().on_training_batch_end(state, should_log)

        if should_log:
            root_logger = logging.get_root_logger()
            tb_logger = hlc.get_from_optional(logging.get_tensorboard_writer())

            loss_val = self._loss_items["loss"]

            root_logger.info(
                f"[{state.global_epoch}, {state.global_iteration:5d}] "
                f"loss {loss_val:.3f}, "
                f"running loss: {loss_val / state.running_iter:.3f} "
                f"avg time: {state.average_iter_time:.2f}s"
            )
            tb_logger.add_scalar("train/loss", loss_val, state.global_iteration)
            tb_logger.add_scalar(
                "train/running loss",
                loss_val / state.running_iter,
                state.global_iteration,
            )

    def on_training_end(self) -> None:
        total = self._val_scores["total"]
        correct = self._val_scores["correct"]
        accuracy = 100 * correct / total
        tb_logger = hlc.get_from_optional(logging.get_tensorboard_writer())
        tb_logger.add_hparams(
            self._tune_params,
            {"hparam/accuracy": accuracy, "hparam/loss": self._loss_items["loss"].item()},
        )

        if self.is_distributed and self.rank == 0:
            assert self.trainer.queue is not None
            self.trainer.queue.put(
                {"accuracy": accuracy, "loss": self._loss_items["loss"].item()}
            )

    def eval(self) -> None:
        self._net.eval()

    def on_validation_start(self, validation_cycle: int) -> None:
        super().on_validation_start(validation_cycle)

        self._val_scores["total"] = 0
        self._val_scores["correct"] = 0
        self._val_scores["loss"] = 0
        self._val_scores["steps"] = 0

    def valid_step(self, batch: typing.Any, step: int) -> None:
        images: torch.Tensor
        labels: torch.Tensor
        images, labels = batch
        images = images.to(self.device)
        labels = labels.to(self.device)

        outputs = self._net(images)
        _, predicted = torch.max(outputs.data, 1)
        self._val_scores["total"] += labels.size(0)
        self._val_scores["correct"] += (predicted == labels).sum().item()

        loss = self._criterion(outputs, labels)
        self._val_scores["loss"] += loss.cpu().numpy()
        self._val_scores["steps"] += 1

    def on_validation_end(self, validation_cycle: int) -> None:
        root_logger = logging.get_root_logger()
        tb_logger = hlc.get_from_optional(logging.get_tensorboard_writer())

        total = self._val_scores["total"]
        correct = self._val_scores["correct"]
        accuracy = 100 * correct // total

        root_logger.info(f"[Validation {validation_cycle}] accuracy: {accuracy}")
        tb_logger.add_scalar("validation/accuracy", accuracy, validation_cycle)
        tb_logger.add_scalar(
            "validation/loss",
            self._val_scores["loss"] / self._val_scores["steps"],
            validation_cycle,
        )

        # Report metrics to the trial
        plugin = self.trainer.plugins[OptunaPlugin.plugin_id]
        assert isinstance(plugin, OptunaPlugin)
        plugin.report_metrics(validation_cycle)

    def append_metadata_to_chkpt_name(self, chkpt_name: str) -> str:
        if len(self._val_scores) == 0:
            return chkpt_name

        total = self._val_scores["total"]
        correct = self._val_scores["correct"]
        accuracy = 100 * correct / total
        chkpt_name += f"_accuracy{round(accuracy, 4)}"
        return chkpt_name


def objective(trial: optuna.Trial) -> float:
    batch_size = trial.suggest_categorical("batch_size", [2, 4, 8, 16])

    datamodule = CIFARDataModule(pathlib.Path.cwd(), batch_size)
    model = ClassifierModel(batch_size)
    plugin = OptunaPlugin(trial, "accuracy")

    trainer = hlt.Trainer(
        run_name="cifar10",
        train_unit=hlt.TrainingUnit.EPOCH,
        total_steps=2,
        valid_frequency=1,
        chkpt_frequency=1,
        print_frequency=10,
        enable_tensorboard=True,
        enable_file_logging=True,
        enable_progress_bar=True,
        enable_deterministic=True,
        print_banner=True,
        chkpt_root=pathlib.Path.cwd() / "chkpt",
        log_path=pathlib.Path.cwd() / "logs",
        run_path=pathlib.Path.cwd() / "runs",
    )

    plugin.configure_trainer(trainer)
    plugin.configure_model(model)
    trainer.fit(model, datamodule)
    plugin.check_pruned()

    if trainer.queue is None:
        return model.metrics["accuracy"]

    metrics = trainer.queue.get()
    return metrics["accuracy"]


if __name__ == "__main__":
    # Set the CUBLAS workspace setting to allow determinism to be used in CUDA >= 10.2.
    os.environ["CUBLAS_WORKSPACE_CONFIG"] = ":4096:8"

    study = optuna.create_study(
        direction="maximize",
        study_name="classifier",
        storage="sqlite:///classifier.db",
        load_if_exists=True,
    )
    study.optimize(objective, n_trials=10, timeout=600)

    pruned_trials = study.get_trials(deepcopy=False, states=[TrialState.PRUNED])
    complete_trials = study.get_trials(deepcopy=False, states=[TrialState.COMPLETE])

    print("Study statistics: ")
    print(f"  Number of finished trials: {len(study.trials)}")
    print(f"  Number of pruned trials: {len(pruned_trials)}")
    print(f"  Number of complete trials: {len(complete_trials)}")

    print("Best trial: ")
    trial = study.best_trial

    print(f"  Value: {trial.value}")
    print("  Params:")
    for key, value in trial.params.items():
        print(f"    {key}: {value}")



from __future__ import annotations

import random
import typing
from collections import abc

import torch
from numpy import random as npr

_DEFAULT_RNG_SEED = 6691


def get_default_seed() -> int:
    """Return the default seed."""
    return _DEFAULT_RNG_SEED


class DefaultNumpyRNG:
    """
    Default RNG from Numpy.

    This is intended to serve as a replacement for the legacy random API from Numpy. The
    class wraps the new Generator instance, which is set to use PCG64. Functionality
    similar to the modules for PyTorch is provided for easy serialization and restoring.

    Args:
        seed: (optional) the initial seed to use.
    """

    def __init__(self, seed: int | list[int] | tuple[int] | None = None):
        """Create the default RNG."""
        self._generator = npr.default_rng(seed)

    @property
    def generator(self) -> npr.Generator:
        """Return the Numpy Generator instance."""
        return self._generator

    def state_dict(self) -> abc.Mapping[str, typing.Any]:
        """
        Create a dictionary containing the RNG state.

        Returns:
            The state of the RNG.
        """
        return self._generator.bit_generator.state

    def load_state_dict(self, state_dict: abc.Mapping[str, typing.Any]) -> None:
        """
        Restore the RNG from the given state dictionary.

        Args:
            state_dict: the state dictionary.
        """
        self._generator.bit_generator.state = state_dict


_DEFAULT_RNG: DefaultNumpyRNG | None = None


def _get_safe_default_rng() -> DefaultNumpyRNG:
    global _DEFAULT_RNG
    if _DEFAULT_RNG is None:
        raise RuntimeError(
            "error: default RNG has not been created. Did you forget to call "
            "create_default_rng?"
        )
    return _DEFAULT_RNG


def create_default_numpy_rng(seed: int | list[int] | tuple[int] | None = None):
    """
    Initialize the default RNG with the given seed.

    Args:
        seed: (optional) the seed to use.
    """
    global _DEFAULT_RNG
    _DEFAULT_RNG = DefaultNumpyRNG(seed=seed)


def get_default_numpy_rng() -> DefaultNumpyRNG:
    """
    Return the default RNG.

    Return:
        The random generator.

    Raises:
        RuntimeError: if the default Numpy RNG hasn't been created.
    """
    return _get_safe_default_rng()


def seed_rngs(seed: int | None = None, skip_torch: bool = False) -> None:
    """
    Seed the default RNGs with the given seed.

    If no seed is given, then the default seed from Helios will be used. The RNGs that
    will be seeded are: PyTorch (+ CUDA if available), stdlib random, and the default
    Numpy generator.
    The ``skip_torch`` flag is intended to be used when seeding worker processes for
    dataloaders. In those cases, the RNGs for PyTorch have already been seeded, so we
    shouldn't be re-seeding them.

    Args:
        seed: optional value to seed the random generators with.
        skip_torch: if True, torch RNGs won't be seeded.
    """
    seed = get_default_seed() if seed is None else seed

    if not skip_torch:
        torch.manual_seed(seed)
        if torch.cuda.is_available():
            torch.cuda.manual_seed_all(seed)
    random.seed(seed)
    create_default_numpy_rng(seed)


def get_rng_state_dict() -> dict[str, typing.Any]:
    """
    Get the state dict for the default RNGs.

    Default RNGs are: PyTorch (+ CUDA if available) and Random.

    Returns:
        The state of all RNGs.
    """
    state = {
        "torch": torch.get_rng_state(),
        "rand": random.getstate(),
        "numpy": get_default_numpy_rng().state_dict(),
    }

    if torch.cuda.is_available():
        state["cuda"] = torch.cuda.get_rng_state()

    return state


def load_rng_state_dict(state_dict: dict[str, typing.Any]) -> None:
    """
    Restore the default RNGs from the given state dict.

    See :py:func:`.get_rng_state_dict` for the list of default RNGs.

    Args:
        state_dict: the state of the RNGs
    """
    torch.set_rng_state(state_dict["torch"])
    random.setstate(state_dict["rand"])
    get_default_numpy_rng().load_state_dict(state_dict["numpy"])
    if torch.cuda.is_available() and "cuda" in state_dict:
        torch.cuda.set_rng_state(state_dict["cuda"])



import torch

from helios import losses


@losses.LOSS_REGISTRY.register
class SampleLoss(losses.WeightedLoss):
    def __init__(self):
        super().__init__(loss_weight=0.5)

    def _eval(self, x: torch.Tensor) -> torch.Tensor:
        return x


class TestLosses:
    def test_registry(self, check_registry) -> None:
        check_registry(
            losses.LOSS_REGISTRY,
            [
                # PyTorch losses
                "L1Loss",
                "MSELoss",
                "CrossEntropyLoss",
                "CTCLoss",
                "NLLLoss",
                "PoissonNLLLoss",
                "GaussianNLLLoss",
                "KLDivLoss",
                "BCELoss",
                "BCEWithLogitsLoss",
                "MarginRankingLoss",
                "HingeEmbeddingLoss",
                "MultiLabelMarginLoss",
                "HuberLoss",
                "SmoothL1Loss",
                "SoftMarginLoss",
                "MultiLabelMarginLoss",
                "CosineEmbeddingLoss",
                "MultiMarginLoss",
                "TripletMarginLoss",
                "TripletMarginWithDistanceLoss",
            ],
        )

    def test_create(self, check_create_function) -> None:
        check_create_function(losses.LOSS_REGISTRY, losses.create_loss)

    def test_weighted_loss(self) -> None:
        loss = losses.create_loss("SampleLoss")
        assert isinstance(loss, SampleLoss)

        x = torch.tensor(10)
        y = loss(x)
        assert y == (x * 0.5)



import pathlib

import torch
from torch import nn

from helios import onnx


class MLPModel(nn.Module):
    def __init__(self):
        super().__init__()
        self.fc0 = nn.Linear(8, 8, bias=True)
        self.fc1 = nn.Linear(8, 4, bias=True)
        self.fc2 = nn.Linear(4, 2, bias=True)
        self.fc3 = nn.Linear(2, 2, bias=True)

    def forward(self, tensor_x: torch.Tensor):
        tensor_x = self.fc0(tensor_x)
        tensor_x = torch.sigmoid(tensor_x)
        tensor_x = self.fc1(tensor_x)
        tensor_x = torch.sigmoid(tensor_x)
        tensor_x = self.fc2(tensor_x)
        tensor_x = torch.sigmoid(tensor_x)
        output = self.fc3(tensor_x)
        return output


class TestONNX:
    def test_jit(self, tmp_path: pathlib.Path) -> None:
        model = MLPModel()
        tensor_x = torch.rand((97, 8), dtype=torch.float32)
        out_path = tmp_path / "mlp.onnx"
        onnx.export_to_onnx(model, tensor_x, out_path, validate_output=True)
        assert out_path.exists()



import pathlib

import numpy as np
import onnx
import onnxruntime  # type: ignore[import-untyped]
import torch
import torch.onnx
from torch import nn


def export_to_onnx(
    net: nn.Module,
    net_args: torch.Tensor,
    out_path: pathlib.Path,
    validate_output: bool = False,
    save_on_validation_fail: bool = True,
    rtol: float = 1e-3,
    atol: float = 1e-5,
    **kwargs,
) -> None:
    """
    Export the given network to ONNX format.

    By default, the resulting onnx network will be validated through ONNX to ensure it's
    valid. If you wish to validate the traced outputs to ensure they're the same, set
    ``validate_output`` to true and change ``rtol``/``atol`` as needed.

    .. warning::
        This function assumes that the given network takes a single tensor as input and
        produces a single tensor as output.

    Args:
        net: the network to convert.
        net_args: the input tensor for tracing.
        out_path: the path to save the exported network to.
        validate_output: if true, validation is performed to ensure correctness. Defaults
            to false.
        save_on_validation_fail: if true, the ONNX network is saved regardless of whether
            validation succeeds. If false, the ONNX network is deleted. Defaults to true.
        rtol: relative tolerance threshold. Defaults to ``1e-3``.
        atol: absolute tolerance threshold. Defaults to ``1e-5``.
        kwargs: additional keyword arguments to ``torch.onnx.export``.
    """
    net.eval()
    with torch.no_grad():
        out = net(net_args)

    torch.onnx.export(net, net_args, out_path, **kwargs)  # type: ignore[arg-type]

    onnx.checker.check_model(out_path)

    if validate_output:
        ort_session = onnxruntime.InferenceSession(
            out_path, providers=["CPUExecutionProvider"]
        )

        def to_numpy(tensor: torch.Tensor):
            if tensor.requires_grad:
                return tensor.detach().cpu().numpy()
            return tensor.cpu().numpy()

        ort_inputs = {ort_session.get_inputs()[0].name: to_numpy(net_args)}
        ort_outs = ort_session.run(None, ort_inputs)
        try:
            np.testing.assert_allclose(to_numpy(out), ort_outs[0], rtol=rtol, atol=atol)
        except AssertionError as e:
            if not save_on_validation_fail:
                out_path.unlink(missing_ok=True)
            raise e



from helios import optim


class TestOptimizers:
    def test_registry(self, check_registry) -> None:
        check_registry(
            optim.OPTIMIZER_REGISTRY,
            [
                "Adadelta",
                "Adagrad",
                "Adam",
                "AdamW",
                "SparseAdam",
                "Adamax",
                "ASGD",
                "LBFGS",
                "NAdam",
                "RAdam",
                "RMSprop",
                "Rprop",
                "SGD",
            ],
        )

    def test_create(self, check_create_function) -> None:
        check_create_function(optim.OPTIMIZER_REGISTRY, optim.create_optimizer)



import math

import torch

from helios import metrics
from helios.core import rng
from helios.metrics import functional


class TestMetrics:
    def test_registry(self, check_registry) -> None:
        check_registry(
            metrics.METRICS_REGISTRY,
            [
                "CalculatePSNR",
                "CalculateSSIM",
                "CalculateMAP",
                "CalculateMAE",
            ],
        )

    def test_create(self, check_create_function) -> None:
        check_create_function(metrics.METRICS_REGISTRY, metrics.create_metric)

    def check_almost_equal(self, val1: float, val2: float) -> None:
        assert math.isclose(val1, val2, abs_tol=0.00001)

    def check_psnr_ssim(self, metric: type, metric_fun_torch, metric_fun) -> None:
        rng.seed_rngs()
        img = torch.rand((1, 3, 32, 32)) * 255
        img2 = torch.rand((1, 3, 32, 32)) * 255
        module = metric(1, "CHW")

        self.check_almost_equal(module(img, img2), metric_fun_torch(img, img2, 1))

        # Convert to numpy
        img = torch.squeeze(img).numpy()  # type: ignore[assignment]
        img2 = torch.squeeze(img2).numpy()  # type: ignore[assignment]

        self.check_almost_equal(module(img, img2), metric_fun(img, img2, 1, "CHW"))

    def test_psnr(self) -> None:
        self.check_psnr_ssim(
            metrics.CalculatePSNR,
            functional.calculate_psnr_torch,
            functional.calculate_psnr,
        )

    def test_ssim(self) -> None:
        self.check_psnr_ssim(
            metrics.CalculateSSIM,
            functional.calculate_ssim_torch,
            functional.calculate_ssim,
        )

    def test_mAP(self) -> None:
        rng.seed_rngs()
        gen = rng.get_default_numpy_rng().generator
        targs = gen.random((1, 10))
        preds = gen.random((1, 10))

        mAP = metrics.CalculateMAP()
        self.check_almost_equal(mAP(targs, preds), functional.calculate_mAP(targs, preds))

    def check_mae(self, scale: float) -> None:
        rng.seed_rngs()

        pred = torch.randn((1, 3, 32, 32)) * scale
        gt = torch.randn((1, 3, 32, 32)) * scale
        mae = metrics.CalculateMAE(scale)

        self.check_almost_equal(
            mae(pred, gt), functional.calculate_mae_torch(pred, gt, scale)
        )

        pred = pred.numpy()  # type: ignore[assignment]
        gt = gt.numpy()  # type: ignore[assignment]
        self.check_almost_equal(mae(pred, gt), functional.calculate_mae(pred, gt, scale))  # type: ignore[arg-type]

    def test_MAE(self) -> None:
        self.check_mae(1.0)
        self.check_mae(255.0)



import abc
import typing

from torch import nn


class WeightedLoss(nn.Module, metaclass=abc.ABCMeta):
    """
    Defines a base class for weighted losses.

    The value of the final loss is determined by the following formula:

        .. math:: L_w = w * L

    where :math:`w` is the weight and :math:`L` is the loss function.

    Example:
        .. code-block:: python

            class MyLoss(WeightedLoss):
                ...
                def _eval(self, ...):
                    return my_loss_function(...)

    Args:
        loss_weight: the weight of the loss function. Defaults to 1.
    """

    def __init__(self, loss_weight: float = 1.0):
        """Create the weighted loss."""
        super().__init__()
        self._loss_weight = loss_weight

    @abc.abstractmethod
    def _eval(self, *args: typing.Any, **kwargs: typing.Any) -> typing.Any:
        """
        Evaluate the loss function.

        Args:
            *args: arguments to the loss function.
            **kwargs: keyword arguments.

        Returns:
            The result of the loss function.
        """

    def forward(self, *args: typing.Any, **kwargs: typing.Any) -> typing.Any:
        """
        Forward wrapper function.

        The final loss value will be computed as described above.

        Args:
            *args: arguments to the loss function.
            **kwargs: keyword arguments.

        Returns:
            The weighted value of the loss function.
        """
        return self._loss_weight * self._eval(*args, **kwargs)



from __future__ import annotations

import math
import typing

import torch
import torch.utils.data as tud

from helios import core

SAMPLER_REGISTRY = core.Registry("sampler")
"""
Global instance of the registry for samplers.

Example:
    .. code-block:: python

        import helios.data.samplers as hlds

        # This automatically registers your sampler.
        @hlds.SAMPLER_REGISTRY.register
        class MySampler:
            ...

        # Alternatively you can manually register a sampler like this:
        hlds.SAMPLER_REGISTRY.register(MySampler)
"""


def create_sampler(
    type_name: str, *args: typing.Any, **kwargs: typing.Any
) -> ResumableSamplerType:
    """
    Create a sampler of the given type.

    This uses the SAMPLER_REGISTRY to look-up sampler types, so ensure your samplers have
    been registered before using this function.

    Args:
        type_name: the type of the transform to create.
        args: positional arguments to pass into the sampler.
        kwargs: keyword arguments to pass into the sampler.

    Returns:
        The constructed sampler.
    """
    return SAMPLER_REGISTRY.get(type_name)(*args, **kwargs)


class ResumableSampler(tud.Sampler):
    r"""
    Base class for samplers that are resumable.

    Let :math:`b_i` be the ith batch for a given epoch :math:`e`. Let the sequence of
    batches that follow be :math:`b_{i + 1}, b_{i + 2}, \ldots`. Suppose that on iteration
    :math:`i`, batch :math:`b_i` is loaded, and training is stopped immediately after. A
    sampler is defined to be resumable if and only if:

    #. Upon re-starting training on epoch :math:`e`, the next batch the sampler loads is
       :math:`b_{i + 1}`.
    #. The order of the subsequent batches :math:`b_{i + 2}, \ldots` must be *identical*
       to the order that the sampler would've produced for the epoch :math:`e` had
       training not stopped.

    Args:
        batch_size: the number of samples per batch.
    """

    def __init__(self, batch_size: int) -> None:
        """Create the sampler."""
        super().__init__()

        self._start_iter: int = 0
        self._epoch: int = 0
        self._batch_size = batch_size

    def _adjust_to_start_iter(self, indices: list[int]) -> list[int]:
        assert self._batch_size > 0

        start_index = self._start_iter * self._batch_size
        return indices[start_index:]

    @property
    def start_iter(self) -> int:
        """The starting iteration for the sampler."""
        return self._start_iter

    @start_iter.setter
    def start_iter(self, ite: int) -> None:
        self._start_iter = ite

    def set_epoch(self, epoch: int) -> None:
        """Set the current epoch for seeding."""
        self._epoch = epoch


@SAMPLER_REGISTRY.register
class ResumableRandomSampler(ResumableSampler):
    """
    Random sampler with resumable state.

    This allows training to stop and resume while guaranteeing that the order in which the
    batches will be returned stays consistent. It is effectively a replacement to the
    default ``RandomSampler`` from PyTorch.

    Args:
        data_source: the dataset to sample from.
        seed: the seed to use for setting up the random generator.
        batch_size: the number of samples per batch.
    """

    def __init__(
        self, data_source: typing.Sized, seed: int = 0, batch_size: int = 1
    ) -> None:
        """Create the sampler."""
        super().__init__(batch_size)

        self._data_source = data_source
        self._seed = seed
        self._num_samples = len(self._data_source)

    def __len__(self) -> int:
        """Return the length of the dataset."""
        return self._num_samples

    def __iter__(self) -> typing.Iterator[int]:
        """Retrieve the index of the next sample."""
        g = torch.Generator()
        g.manual_seed(self._epoch + self._seed)
        shuffling = torch.randperm(self._num_samples, generator=g)
        indices = shuffling.tolist()

        assert len(indices) == self._num_samples
        indices = self._adjust_to_start_iter(indices)
        return iter(indices)


@SAMPLER_REGISTRY.register
class ResumableSequentialSampler(ResumableSampler):
    """
    Sequential sampler with resumable state.

    This allows training to stop and resume while guaranteeing that the order in which the
    batches will be returned stays consistent. It is effectively a replacement to the
    default ``SequentialSampler`` from PyTorch.

    Args:
        data_source: the dataset to sample from.
        batch_size: the number of samples per batch.
    """

    def __init__(self, data_source: typing.Sized, batch_size: int = 1):
        """Create the sampler."""
        super().__init__(batch_size)

        self._data_source = data_source
        self._num_samples = len(data_source)
        self._indices = list(range(self._num_samples))

    def __len__(self) -> int:
        """Return the length of the dataset."""
        return self._num_samples

    def __iter__(self) -> typing.Iterator[int]:
        """Retrieve the index of the next sample."""
        assert self._batch_size > 0
        indices = self._indices
        indices = self._adjust_to_start_iter(indices)
        return iter(indices)


@SAMPLER_REGISTRY.register
class ResumableDistributedSampler(tud.DistributedSampler):
    """
    Distributed sampler with resumable state.

    This allows training to stop and resume while guaranteeing that the order in which the
    batches will be returned stays consistent. It is effectively a replacement to the
    default ``DistributedSampler`` from PyTorch.

    Args:
        dataset: the dataset to sample from.
        num_replicas: number of processes for distributed training.
        rank: (optional) rank of the current process.
        shuffle: if true, shuffle the indices. Defaults to true.
        seed: random seed used to shuffle the sampler. Defaults to 0.
        drop_last: if true, then drop the final sample to make it even across replicas.
            Defaults to false.
        batch_size: the number of samples per batch. Defaults to 1.
    """

    def __init__(
        self,
        dataset: tud.Dataset,
        num_replicas: int | None = None,
        rank: int | None = None,
        shuffle: bool = True,
        seed: int = 0,
        drop_last: bool = False,
        batch_size: int = 1,
    ):
        """Create the sampler."""
        super().__init__(dataset, num_replicas, rank, shuffle, seed, drop_last)
        self._batch_size = batch_size
        self._start_iter = 0

    @property
    def start_iter(self) -> int:
        """The starting iteration for the sampler."""
        return self._start_iter

    @start_iter.setter
    def start_iter(self, ite: int) -> None:
        self._start_iter = ite

    def __iter__(self) -> typing.Iterator[int]:
        """Retrieve the index of the next sample."""
        # The code for this function was adapted from PyTorch's implementation of
        # DistributedSampler. The original license from PyTorch can be viewed here:
        # https://github.com/pytorch/pytorch/blob/main/LICENSE
        if self.shuffle:
            # deterministically shuffle based on epoch and seed
            g = torch.Generator()
            g.manual_seed(self.seed + self.epoch)
            indices = torch.randperm(len(self.dataset), generator=g).tolist()  # type: ignore[arg-type]
        else:
            indices = list(range(len(self.dataset)))  # type: ignore[arg-type]

        if not self.drop_last:
            # add extra samples to make it evenly divisible
            padding_size = self.total_size - len(indices)
            if padding_size <= len(indices):
                indices += indices[:padding_size]
            else:
                indices += (indices * math.ceil(padding_size / len(indices)))[
                    :padding_size
                ]
        else:
            # remove tail of data to make it evenly divisible.
            indices = indices[: self.total_size]
        assert len(indices) == self.total_size

        # subsample
        indices = indices[self.rank : self.total_size : self.num_replicas]
        assert len(indices) == self.num_samples

        # Trim off samples that we have already seen if we're restarting.
        assert self._batch_size > 0
        start_index = self._start_iter * self._batch_size
        indices = indices[start_index:]

        return iter(indices)


ResumableSamplerType = ResumableSampler | ResumableDistributedSampler
"""
Defines the resumable sampler type.

A resumable sampler **must** be derived from either
:py:class:`~helios.data.samplers.ResumableSampler` or
:py:class:`~helios.data.samplers.ResumableDistributedSampler`.
"""



from helios import scheduler


class TestSchedulers:
    def test_registry(self, check_registry) -> None:
        check_registry(
            scheduler.SCHEDULER_REGISTRY,
            [
                # Pytorch schedulers.
                "LambdaLR",
                "MultiplicativeLR",
                "StepLR",
                "MultiStepLR",
                "ConstantLR",
                "LinearLR",
                "ExponentialLR",
                "PolynomialLR",
                "CosineAnnealingLR",
                "SequentialLR",
                "ReduceLROnPlateau",
                "CyclicLR",
                "OneCycleLR",
                "CosineAnnealingWarmRestarts",
                # Custom schedulers.
                "MultiStepRestartLR",
                "CosineAnnealingRestartLR",
            ],
        )

    def test_create(self, check_create_function) -> None:
        check_create_function(scheduler.SCHEDULER_REGISTRY, scheduler.create_scheduler)



"""
Scheduler package for Helios.

This contains the classes, utilities, and registries related to schedulers.
"""

from .schedulers import CosineAnnealingRestartLR, MultiStepRestartLR
from .utils import SCHEDULER_REGISTRY, create_scheduler

__all__ = [
    "CosineAnnealingRestartLR",
    "MultiStepRestartLR",
    "SCHEDULER_REGISTRY",
    "create_scheduler",
]



import dataclasses as dc
import functools
import pathlib
import typing

import optuna
import pytest
import torch

import helios.model as hlm
import helios.plugins as hlp
import helios.plugins.optuna as hlpo
import helios.trainer as hlt
from helios.core import rng

# Ignore the use of private members so we can test them correctly.
# ruff: noqa: SLF001


class ExceptionPlugin(hlp.Plugin):
    def __init__(self, exc_type: type[Exception] | list[type[Exception]]):
        super().__init__(f"exception_{str(exc_type)}")
        self._exc_type = exc_type

    def setup(self) -> None:
        pass

    def configure_trainer(self, t: hlt.Trainer) -> None:
        self._append_train_exceptions(self._exc_type, t)
        self._append_test_exceptions(self._exc_type, t)


class PluginModel(hlm.Model):
    def __init__(self, save_name: str):
        super().__init__(save_name)

    def setup(self, fast_init: bool = False) -> None:
        pass


class TestPlugins:
    def test_registry(self, check_registry) -> None:
        check_registry(hlp.PLUGIN_REGISTRY, ["CUDAPlugin", "OptunaPlugin"])

    def test_create(self, check_create_function) -> None:
        check_create_function(hlp.PLUGIN_REGISTRY, hlp.create_plugin)

    def test_append_exceptions(self) -> None:
        t = hlt.Trainer()
        plugin = ExceptionPlugin(RuntimeError)
        exc_list: list[type[Exception]] = [RuntimeError]

        plugin.configure_trainer(t)
        assert t.train_exceptions == exc_list
        assert t.test_exceptions == exc_list

        exc_list.extend([ValueError, TypeError])
        plugin = ExceptionPlugin([ValueError, TypeError])
        plugin.configure_trainer(t)

        assert t.train_exceptions == exc_list
        assert t.test_exceptions == exc_list


class TestCUDAPlugin:
    def check_batch_device(self, x: typing.Any, device: torch.device) -> None:
        if isinstance(x, torch.Tensor):
            assert x.device == device
        elif isinstance(x, dict):
            assert all(val.device == device for _, val in x.items())
        else:
            assert all(elem.device == device for elem in x)

    def check_batch_processing(
        self, plugin: hlp.CUDAPlugin, x: typing.Any, device: torch.device
    ) -> None:
        ret = plugin.process_training_batch(x, hlt.TrainingState())
        self.check_batch_device(ret, device)

        ret = plugin.process_validation_batch(x, 0)
        self.check_batch_device(ret, device)

        ret = plugin.process_testing_batch(x, 0)
        self.check_batch_device(ret, device)

    def test_plugin_id(self) -> None:
        assert hasattr(hlp.CUDAPlugin, "plugin_id")
        assert hlp.CUDAPlugin.plugin_id == "cuda"

    def test_batch_processing(self) -> None:
        if not torch.cuda.is_available():
            return

        def create_tensor() -> torch.Tensor:
            return torch.randn((1, 3, 32, 32)).to("cpu")

        device = torch.device("cuda:0")

        plugin = hlp.CUDAPlugin()
        plugin.is_distributed = False
        plugin.map_loc = {"cuda:0": "cuda:0"}
        plugin.device = device

        assert plugin.unique_overrides.training_batch
        assert plugin.unique_overrides.validation_batch
        assert plugin.unique_overrides.testing_batch

        self.check_batch_processing(plugin, create_tensor(), device)
        self.check_batch_processing(plugin, [create_tensor(), create_tensor()], device)
        self.check_batch_processing(plugin, (create_tensor(), create_tensor()), device)
        self.check_batch_processing(
            plugin, {"a": create_tensor(), "b": create_tensor()}, device
        )

    def test_configure(self) -> None:
        if not torch.cuda.is_available():
            return

        trainer = hlt.Trainer()
        plugin = hlp.CUDAPlugin()
        plugin.configure_trainer(trainer)

        assert "cuda" in trainer.plugins
        assert trainer.plugins["cuda"] == plugin


# Ignore the warnings coming from optuna.
@pytest.mark.filterwarnings(
    ("ignore::optuna.exceptions.ExperimentalWarning"), ("ignore::FutureWarning")
)
class TestOptunaPlugin:
    def test_plugin_id(self) -> None:
        assert hasattr(hlpo.OptunaPlugin, "plugin_id")
        assert hlpo.OptunaPlugin.plugin_id == "optuna"

    def test_invalid_storage(self) -> None:
        def objective(trial: optuna.Trial) -> int:
            plugin = hlpo.OptunaPlugin(trial, "accuracy")
            plugin.is_distributed = True
            with pytest.raises(ValueError):
                plugin.setup()

            return 0

        study = optuna.create_study()
        study.optimize(objective, n_trials=1)

    def test_configure(self) -> None:
        def objective(trial: optuna.Trial) -> int:
            plugin = hlpo.OptunaPlugin(trial, "accuracy")
            trainer = hlt.Trainer()

            plugin.configure_trainer(trainer)
            assert len(trainer.plugins) == 1
            assert trainer.plugins["optuna"] == plugin
            assert len(trainer.train_exceptions) == 1
            assert trainer.train_exceptions[0] == optuna.TrialPruned

            model = PluginModel("plugin-model")
            plugin.configure_model(model)
            assert model.save_name == "plugin-model_trial-0"

            return 0

        study = optuna.create_study()
        study.optimize(objective, n_trials=1)

    def check_in_range(
        self, val: typing.Any, t: type, low: typing.Any, high: typing.Any
    ) -> None:
        assert isinstance(val, t)
        assert low <= val <= high

    def check_in_sequence(self, val: typing.Any, t: type, seq: typing.Container) -> None:
        assert isinstance(val, t)
        assert val in seq

    def test_suggest(self) -> None:
        def objective(trial: optuna.Trial) -> int:
            plugin = hlpo.OptunaPlugin(trial, "accuracy")

            with pytest.raises(KeyError):
                plugin.suggest("foo", "bar")

            seq = [1, 2, 3]
            self.check_in_sequence(
                plugin.suggest("categorical", "val1", choices=seq), int, seq
            )

            low = 0
            high = 10
            self.check_in_range(
                plugin.suggest("int", "val2", low=low, high=high), int, low, high
            )

            high = 1
            self.check_in_range(
                plugin.suggest("float", "val3", low=low, high=high), float, low, high
            )

            return 0

        study = optuna.create_study()
        study.optimize(objective, n_trials=1)

    def test_state_dict(self) -> None:
        def objective(trial: optuna.Trial) -> int:
            plugin = hlpo.OptunaPlugin(trial, "accuracy")
            x = plugin.suggest("float", "x", low=-10, high=10)

            state_dict = plugin.state_dict()
            assert len(state_dict) == 1
            assert "x" in state_dict
            assert state_dict["x"] == x

            return 0

        study = optuna.create_study()
        study.optimize(objective, n_trials=1)

    def test_resume_trial(self, tmp_path: pathlib.Path) -> None:
        num_trials = 10
        storage_path = tmp_path / "trial_test.db"
        successful_trials = [False for _ in range(num_trials)]
        offset = 0
        study_args = {
            "study_name": "trial_test",
            "storage": f"sqlite:///{storage_path}",
            "load_if_exists": True,
        }

        def objective(trial: optuna.Trial) -> float:
            nonlocal offset
            plugin = hlpo.OptunaPlugin(trial, "accuracy")
            plugin.setup()

            model = PluginModel("plugin-model")
            plugin.configure_model(model)

            trial_num = trial.number
            if hlpo._ORIG_NUMBER_KEY in trial.user_attrs:
                trial_num = trial.user_attrs[hlpo._ORIG_NUMBER_KEY]
                assert model.save_name == f"plugin-model_trial-{trial_num}"

                # Artificially offset the trial number so we don't raise the exception
                # again.
                trial_num += 1
                offset = 1

            if trial_num == num_trials / 2:
                raise RuntimeError("half-way stop")

            successful_trials[trial_num - offset] = True
            offset = 0
            if trial_num % 2 == 1:
                raise optuna.TrialPruned()

            return 0

        def optimize(study: optuna.Study) -> None:
            study.optimize(
                objective,
                n_trials=num_trials,
                callbacks=[
                    optuna.study.MaxTrialsCallback(
                        num_trials,
                        states=(
                            optuna.trial.TrialState.COMPLETE,
                            optuna.trial.TrialState.PRUNED,
                        ),
                    )
                ],
            )

        study = hlpo.resume_study(study_args, backup_study=False)
        assert len(study.trials) == 0
        with pytest.raises(RuntimeError):
            optimize(study)

        del study

        study = hlpo.resume_study(study_args, backup_study=False)
        assert len(study.trials) == (num_trials // 2) + 1
        optimize(study)

        for v in successful_trials:
            assert v

    def test_study_backup(self, tmp_path: pathlib.Path) -> None:
        storage_path = tmp_path / "test.db"
        storage_path.touch(exist_ok=True)

        for _ in range(10):
            hlpo._backup_study(storage_path)

        # Ensure there are 11 "trials"
        backups = list(tmp_path.glob("*.db"))
        assert len(backups) == 11

        idx = [i for i, b in enumerate(backups) if b.name == "test.db"][0]
        backups.pop(idx)

        value = [False] * 10
        for p in backups:
            idx = int(p.stem.split("-")[-1])
            value[idx] = True

        assert all(v for v in value)

    def test_sampler_checkpoints(self, tmp_path: pathlib.Path) -> None:
        @dc.dataclass
        class TestRun:
            samples: list[float]
            chkpt_root: pathlib.Path

        num_trials = 10
        run1 = TestRun([], tmp_path / "run1")
        run2 = TestRun([], tmp_path / "run2")

        run1.chkpt_root.mkdir(exist_ok=True)
        run2.chkpt_root.mkdir(exist_ok=True)

        def objective(
            trial: optuna.Trial,
            raise_error: bool,
            run: TestRun,
        ) -> float:
            hlpo.checkpoint_sampler(trial, run.chkpt_root)
            res = trial.suggest_float("accuracy", 0, 1)
            if raise_error and trial.number == num_trials // 2:
                raise RuntimeError("half-way stop")
            run.samples.append(res)
            return res

        def create_study(
            sampler: optuna.samplers.BaseSampler | None = None,
        ) -> optuna.Study:
            return optuna.create_study(
                study_name="chkpt_test",
                sampler=optuna.samplers.TPESampler(seed=rng.get_default_seed())
                if sampler is None
                else sampler,
            )

        study = create_study()
        study.optimize(
            functools.partial(
                objective,
                raise_error=False,
                run=run1,
            ),
            n_trials=num_trials,
        )

        chkpts = [chkpt.stem for chkpt in run1.chkpt_root.glob("*.pkl")]
        chkpts.sort()
        assert len(chkpts) == num_trials
        assert all(chkpt == f"sampler_trial-{i}" for i, chkpt in enumerate(chkpts))
        del study

        study = create_study()
        sampler = hlpo.restore_sampler(run2.chkpt_root)
        assert sampler is None

        with pytest.raises(RuntimeError):
            study.optimize(
                functools.partial(
                    objective,
                    raise_error=True,
                    run=run2,
                ),
                n_trials=num_trials,
            )

        torch.serialization.clear_safe_globals()
        sampler = hlpo.restore_sampler(run2.chkpt_root)
        study = create_study(sampler)
        study.optimize(
            functools.partial(objective, raise_error=False, run=run2),
            n_trials=num_trials // 2,
        )
        assert run2.samples == run1.samples



import typing

import numpy.typing as npt
import PIL
import torch
import torchvision.transforms.v2 as T
from torch import nn

from helios import core

TRANSFORM_REGISTRY = core.Registry("transform")
"""
Global instance of the registry for transforms.

Example:
    .. code-block:: python

        import helios.data.transforms as hldt

        # This automatically registers your dataset.
        @hldt.TRANSFORM_REGISTRY.register()
        class MyTransform:
            ...

        # Alternatively you can manually register a dataset like this:
        hldt.TRANSFORM_REGISTRY.register(MyTransform)
"""


def create_transform(
    type_name: str, *args: typing.Any, **kwargs: typing.Any
) -> nn.Module:
    """
    Create a transform of the given type.

    This uses TRANSFORM_REGISTRY to look-up transform types, so ensure your transforms
    have been registered before using this function.

    Args:
        type_name: the type of the transform to create.
        args: positional arguments to pass into the transform.
        kwargs: keyword arguments to pass into the transform.

    Returns:
        The constructed transform.
    """
    return TRANSFORM_REGISTRY.get(type_name)(*args, **kwargs)


@TRANSFORM_REGISTRY.register
class ToImageTensor(nn.Module):
    """
    Convert an image (or list of images) to tensor(s).

    An image is meant to be a tensor, ndarray, or PIL image. The shape expected to be
    either [H, W, C] or [C, H, W].

    Args:
        dtype: the output type of the tensors.
        scale: if true, scale the values to the valid range. Defaults to true.
    """

    def __init__(self, dtype: torch.dtype = torch.float32, scale: bool = True):
        """Create the transform."""
        super().__init__()
        self._transform = T.Compose(
            [T.ToImage(), T.ToDtype(dtype, scale=scale), T.ToPureTensor()]
        )

    def forward(
        self,
        img: npt.NDArray
        | list[npt.NDArray]
        | tuple[npt.NDArray, ...]
        | PIL.Image.Image
        | list[PIL.Image.Image]
        | tuple[PIL.Image.Image, ...],
    ) -> torch.Tensor | list[torch.Tensor] | tuple[torch.Tensor, ...]:
        """
        Convert the input image(s) into tensor(s).

        The return type will match the type of the input. So, if the input is a single
        image, then the output will be a single tensor. If the input is a list or a tuple
        of images, the output will be a list or tuple of tensors.

        Args:
            img: image(s) to convert.

        Returns:
            The converted images.
        """
        out_tens: list[torch.Tensor] = []
        for elem in core.convert_to_list(img):
            out_tens.append(self._transform(elem))

        if len(out_tens) == 1:
            return out_tens[0]
        if isinstance(img, tuple):
            return tuple(out_tens)
        return out_tens



import typing

import torch

import helios.model as hlm
import helios.trainer as hlt


@hlm.MODEL_REGISTRY.register
class MockModel(hlm.Model):
    def __init__(self) -> None:
        super().__init__("mock-model")

    @property
    def loss_items(self) -> dict[str, torch.Tensor]:
        return self._loss_items

    @property
    def running_loss(self) -> dict[str, float]:
        return self._running_losses

    @property
    def val_scores(self) -> dict[str, float]:
        return self._val_scores

    @property
    def test_scores(self) -> dict[str, float]:
        return self._test_scores

    def setup(self, fast_init: bool = False) -> None:
        pass

    def train_step(self, batch: typing.Any, state) -> None:
        pass

    def populate_loss(self) -> None:
        self._loss_items["foo"] = torch.tensor(1)
        self._loss_items["bar"] = torch.tensor(2)

    def populate_val_scores(self) -> None:
        self._val_scores["foo"] = 1
        self._val_scores["bar"] = 2

    def populate_test_scores(self) -> None:
        self._test_scores["foo"] = 1
        self._test_scores["bar"] = 2


class TestModel:
    def test_defaults(self) -> None:
        model = MockModel()
        state = hlt.TrainingState()

        assert model.save_name == "mock-model"
        assert model.state_dict() == {}
        assert model.trained_state_dict() == {}

        chkpt_name = "chkpt"
        assert model.append_metadata_to_chkpt_name(chkpt_name) == chkpt_name

        model.on_training_batch_start(state)
        assert len(model.loss_items) == 0
        model.populate_loss()
        assert len(model.loss_items) != 0
        model.on_training_batch_end(state)
        assert len(model.running_loss) != 0
        model.on_training_batch_start(state)
        assert len(model.loss_items) == 0

        model.populate_val_scores()
        assert len(model.val_scores) != 0
        model.on_validation_start(0)
        assert len(model.loss_items) == 0
        model.on_validation_end(0)
        assert len(model.running_loss) == 0

        assert model.have_metrics_improved()

        model.populate_test_scores()
        assert len(model.test_scores) != 0
        model.on_testing_start()
        assert len(model.test_scores) == 0

    def test_create(self, check_registry) -> None:
        check_registry(hlm.MODEL_REGISTRY, ["MockModel"])
        hlm.create_model("MockModel")



from __future__ import annotations

import pathlib
import typing

from helios import core

if typing.TYPE_CHECKING:
    from .model import Model

MODEL_REGISTRY = core.Registry("model")
"""
Global instance of the registry for models.

Example:
    .. code-block:: python

        import helios.model as hlm

        # This automatically registers your model.
        @hlm.MODEL_REGISTRY.register
        class MyModel:
            ...

        # Alternatively you can manually register a model like this:
        hlm.MODEL_REGISTRY.register(MyModel)
"""


def create_model(type_name: str, *args: typing.Any, **kwargs: typing.Any) -> Model:
    """
    Create the model for the given type.

    Args:
        type_name: the type of the model to create.
        args: positional arguments to pass into the model.
        kwargs: keyword arguments to pass into the model.

    Returns:
        The model.
    """
    return MODEL_REGISTRY.get(type_name)(*args, **kwargs)


def find_pretrained_file(root: pathlib.Path, name: str) -> pathlib.Path:
    """
    Find the pre-trained file in the given root.

    The assumption is the following:
    Given a root ``/models/cifar`` and a name ``resnet-50``, then the name of the
    pre-trained file will contain ``cifar_resnet-50_`` as a prefix. If no file is
    found, an exception is raised.

    Args:
        root: the root where the file is stored.
        net_name: the save name of the file.

    Returns:
        The path to the file.

    Raises:
        RuntimeError: if no pre-trained network was found.
    """
    for path in root.glob("*.pth"):
        file_name = str(path.stem)
        base_name = f"{str(root.stem)}_{name}_"
        if base_name in file_name:
            return path

    raise RuntimeError(
        f"error: unable to find a pretrained network named {name} at {str(root)}"
    )



import dataclasses
import typing

import pytest
from torch import nn

from helios import core


class SampleModuleNoArgs(nn.Module):
    def forward(self, x: int) -> int:
        return x


class SampleModuleArgs(nn.Module):
    def __init__(self, val: int):
        super().__init__()

        self._val = val

    def forward(self, x: int) -> int:
        return x * self._val


class SampleModuleKwargs(nn.Module):
    def __init__(self, val: int = 1):
        super().__init__()

        self._val = val

    def forward(self, x: int) -> int:
        return x * self._val


class SampleModuleArgsAndKwargs(nn.Module):
    def __init__(self, val: int, opt_val: int = 1):
        super().__init__()

        self._val = val
        self._opt_val = opt_val

    def forward(self, x: int) -> int:
        return x * self._opt_val + self._val


@dataclasses.dataclass
class SampleEntry:
    sample_type: type
    exp_ret: int
    args: list[typing.Any] = dataclasses.field(default_factory=list)
    kwargs: dict[str, typing.Any] = dataclasses.field(default_factory=dict)


@pytest.fixture
def check_registry() -> typing.Callable[[core.Registry, list[str]], None]:
    def impl_fun(registry: core.Registry, registered_names: list[str]) -> None:
        assert len(typing.cast(typing.Sized, registry.keys())) != 0

        for name in registered_names:
            assert name in registry

    return impl_fun


@pytest.fixture
def check_create_function() -> typing.Callable[[core.Registry, typing.Callable], None]:
    def impl_fun(registry: core.Registry, create_fun: typing.Callable) -> None:
        in_val = 4
        test_table = [
            SampleEntry(SampleModuleNoArgs, exp_ret=in_val),
            SampleEntry(SampleModuleArgs, exp_ret=in_val * 2, args=[2]),
            SampleEntry(SampleModuleKwargs, exp_ret=in_val * 2, kwargs={"val": 2}),
            SampleEntry(
                SampleModuleArgsAndKwargs,
                exp_ret=in_val * 2 + 2,
                args=[2],
                kwargs={"opt_val": 2},
            ),
        ]

        for entry in test_table:
            registry.register(entry.sample_type)

        for entry in test_table:
            ret = create_fun(entry.sample_type.__name__, *entry.args, **entry.kwargs)
            assert isinstance(ret, entry.sample_type)
            val = ret(in_val)  # type: ignore[operator]
            assert val == entry.exp_ret

    return impl_fun



import dataclasses as dc
import os
import typing

import torch
import torch.distributed as dist


def is_using_torchrun() -> bool:
    """
    Check if the current process was launched from ``torchrun``.

    This will check to see if the environment variables that are set by ``torchrun``
    exist. The list of variables is taken directly from the documentation and can be seen
    `here`_.

    .. _here: https://pytorch.org/docs/stable/elastic/run.html#environment-variables

    Returns:
        True if the run was started from ``torchrun``, false otherwise.
    """
    return all(
        key in os.environ
        for key in [
            "LOCAL_RANK",
            "RANK",
            "GROUP_RANK",
            "ROLE_RANK",
            "LOCAL_WORLD_SIZE",
            "WORLD_SIZE",
            "ROLE_WORLD_SIZE",
            "MASTER_ADDR",
            "MASTER_PORT",
            "TORCHELASTIC_RESTART_COUNT",
            "TORCHELASTIC_RUN_ID",
        ]
    )


def init_dist(
    backend: str = "nccl", rank: int | None = None, world_size: int | None = None
) -> None:
    """
    Initialize the distributed process group.

    The optional values for ``rank`` and ``world_size`` **must** be omitted if
    distributed training is handled through ``torchrun``. If distributed training is
    started manually, then both arguments **must** be provided.

    Args:
        backend: the backend to use. Defaults to "nccl".
        rank: the (optional) rank of the current GPU.
        world_size: the (optional) number of GPUs in the system.

    Raises:
        ValueError: if either of ``rank`` or ``world_size`` are ``None`` (but not both).
    """
    if rank is None and world_size is None:
        dist.init_process_group(backend)
        return

    if rank is None or world_size is None:
        raise ValueError("error: rank and world_size cannot be None")

    os.environ["LOCAL_RANK"] = str(rank)
    os.environ["RANK"] = str(rank)
    os.environ["LOCAL_WORLD_SIZE"] = str(world_size)
    os.environ["WORLD_SIZE"] = str(world_size)
    os.environ["MASTER_ADDR"] = "localhost"
    os.environ["MASTER_PORT"] = "12355"

    device: torch.device | None = None
    if backend == "nccl":
        device = torch.device("cuda", rank)

    dist.init_process_group(backend, rank=rank, world_size=world_size, device_id=device)


def shutdown_dist() -> None:
    """Shutdown the distributed process group."""
    dist.destroy_process_group()


def is_dist_active() -> bool:
    """Check if torch.distributed is active."""
    return dist.is_available() and dist.is_initialized()


@dc.dataclass
class DistributedInfo:
    """
    Bundle information regarding distributed state.

    To understand what these mean, consider a run being performed over two nodes, each
    with 2 GPUs. Suppose we have a single process per GPU. Then the following values are
    assigned:

    * ``local_rank``: 0 or 1 for both nodes. 0 for the first GPU, 1 for the second.
    * ``rank``: 0, 1, 2, 3 for each of the GPUs over the two nodes.
    * ``local_world_size``: 2 for both nodes.
    * ``world_size``: 4 (2 nodes with 2 workers per node).

    Args:
        local_rank: the local rank.
        rank: the global rank.
        local_world_size: the local world size.
        world_size: the global world size.
    """

    local_rank: int = 0
    rank: int = 0
    local_world_size: int = 1
    world_size: int = 1


def get_dist_info() -> DistributedInfo:
    """
    Get the distributed state of the current run.

    If distributed training is not used, then both ranks are set to 0 and both world sizes
    are set to 1.

    Returns:
        The information of the current distributed run.
    """
    info = DistributedInfo()
    if dist.is_available() and dist.is_initialized():
        info.local_rank = int(os.environ["LOCAL_RANK"])
        info.rank = int(os.environ["RANK"])
        info.local_world_size = int(os.environ["LOCAL_WORLD_SIZE"])
        info.world_size = int(os.environ["WORLD_SIZE"])

    return info


def get_local_rank() -> int:
    """
    Get the local rank of the device the process is running on.

    If distributed training is not used, 0 is returned.

    Returns:
        The local rank of the current device.
    """
    return get_dist_info().local_rank


def get_global_rank() -> int:
    """
    Get the global rank of the device the process is running on.

    If distributed training is not used, 0 is returned.

    Returns:
        The global rank of the current device.
    """
    return get_dist_info().rank


def gather_into_tensor(tensor: torch.Tensor, size: tuple[int]) -> torch.Tensor:
    """
    Gathers the tensors across all processes and merges them into a single tensor.

    Args:
        tensor: the tensor to merge
        size: the dimensions of the output tensor.

    Returns:
        The resulting tensor containing all gathered tensors.

    Raises:
        RuntimeError: if distributed training hasn't been initialised.
    """
    if not dist.is_initialized():
        raise RuntimeError("error: default process group has not been initialized")

    device = torch.device(f"cuda:{get_local_rank()}")
    out = torch.zeros(size, device=device, dtype=tensor.dtype)
    dist.all_gather_into_tensor(out, tensor)
    return out


def all_reduce_tensors(
    tensor: torch.Tensor | list[torch.Tensor], **kwargs
) -> torch.Tensor:
    """
    Reduces tensors across all processes so all have the same value.

    Args:
        tensor: the input tensor(s) to reduce. If the input is a list of tensors, they
            will be concatenated into a single tensor.
        kwargs: additional options for torch.distributed.all_reduce

    Returns:
        The reduced tensor.

    Raises:
        RuntimeError: if distributed training has not been initialised.
    """
    if not dist.is_initialized():
        raise RuntimeError("error: default process group has not been initialized")

    value_tensor = torch.tensor(tensor).cuda() if isinstance(tensor, list) else tensor
    dist.all_reduce(value_tensor, **kwargs)
    return value_tensor


def _dist_print_wrapper(
    *args: typing.Any, rank_check: typing.Callable[[], bool], **kwargs: typing.Any
) -> None:
    if not is_dist_active():
        print(*args, **kwargs)
        return
    if rank_check():
        print(*args, **kwargs)


def global_print(*args: typing.Any, global_rank: int = 0, **kwargs: typing.Any) -> None:
    """
    Print wrapper that only prints on the specified global rank.

    Args:
        *args: positional arguments to pass in to Python's print.
        global_rank: the global rank the print should happen on. Defaults to 0.
        **kwargs: keyword arguments to pass in to Python's print.
    """
    _dist_print_wrapper(
        *args, rank_check=lambda: get_global_rank() == global_rank, **kwargs
    )


def local_print(*args: typing.Any, local_rank: int = 0, **kwargs: typing.Any) -> None:
    """
    Print wrapper that only prints on the specified local rank.

    Args:
        *args: positional arguments to pass in to Python's print.
        local_rank: the local rank the print should happen on. Defaults to 0.
        **kwargs: keyword arguments to pass in to Python's print.
    """
    _dist_print_wrapper(
        *args, rank_check=lambda: get_local_rank() == local_rank, **kwargs
    )


def safe_barrier(**kwargs: typing.Any) -> None:
    """
    Safe wrapper for torch.distributed.barrier.

    The wrapper is "safe" in the sense that it is valid to call this function regardless
    of whether the code is currently using distributed training or not.

    Args:
        **kwargs: keyword arguments to torch.distributed.barrier.
    """
    if not is_dist_active():
        return
    dist.barrier(**kwargs)



"""
Model package for Helios.

This contains all the classes, utilities, and registries related to models.
"""

from .model import Model
from .utils import MODEL_REGISTRY, create_model, find_pretrained_file

__all__ = ["Model", "MODEL_REGISTRY", "create_model", "find_pretrained_file"]



import cv2
import numpy as np
import numpy.typing as npt
import torch
import torch.nn.functional as F

from helios.data import functional


def _ssim(img: npt.NDArray[np.float64], img2: npt.NDArray[np.float64]) -> float:
    """
    SSIM implementation function. See calculate_ssim for details.

    Args:
        img: Images of type ``np.float64`` with range  :math:`[0, 1]`.
        img2: Images of type ``np.float64`` with range :math:`[0, 1]`.

    Returns:
        SSIM value.
    """
    c1 = (0.01 * 255) ** 2
    c2 = (0.03 * 255) ** 2
    kernel: npt.NDArray = cv2.getGaussianKernel(11, 1.5)
    window = np.outer(kernel, kernel.transpose())

    # valid mode for window size 11
    mu1: npt.NDArray = cv2.filter2D(img, -1, window)[5:-5, 5:-5]
    mu2: npt.NDArray = cv2.filter2D(img2, -1, window)[5:-5, 5:-5]
    mu1_sq = mu1**2
    mu2_sq = mu2**2
    mu1_mu2 = mu1 * mu2
    sigma1_sq = cv2.filter2D(img**2, -1, window)[5:-5, 5:-5] - mu1_sq  # type: ignore[operator]
    sigma2_sq = cv2.filter2D(img2**2, -1, window)[5:-5, 5:-5] - mu2_sq  # type: ignore[operator]
    sigma12 = cv2.filter2D(img * img2, -1, window)[5:-5, 5:-5] - mu1_mu2

    ssim_map = ((2 * mu1_mu2 + c1) * (2 * sigma12 + c2)) / (
        (mu1_sq + mu2_sq + c1) * (sigma1_sq + sigma2_sq + c2)
    )
    return ssim_map.mean()


def _ssim_torch(img: torch.Tensor, img2: torch.Tensor) -> float:
    """
    SSIM Torch implementation function. See :py:func:`.calculate_ssim_torch` for details.

    Args:
        img: Images of type ``torch.float64`` with range  :math:`[0, 1]`.
        img2: Images of type ``torch.float64`` with range :math:`[0, 1]`.

    Returns:
        float: SSIM.
    """
    c1 = (0.01 * 255) ** 2
    c2 = (0.03 * 255) ** 2

    kernel: npt.NDArray = cv2.getGaussianKernel(11, 1.5)
    window = (
        torch.from_numpy(np.outer(kernel, kernel.transpose()))
        .view(1, 1, 11, 11)
        .expand(img.size(1), 1, 11, 11)
        .to(img.dtype)
        .to(img.device)
    )

    # pylint: disable=not-callable
    mu1 = F.conv2d(img, window, stride=1, padding=0, groups=img.shape[1])  # valid mode
    mu2 = F.conv2d(img2, window, stride=1, padding=0, groups=img2.shape[1])  # valid mode
    mu1_sq = mu1.pow(2)
    mu2_sq = mu2.pow(2)
    mu1_mu2 = mu1 * mu2
    sigma1_sq = (
        F.conv2d(img * img, window, stride=1, padding=0, groups=img.shape[1]) - mu1_sq
    )
    sigma2_sq = (
        F.conv2d(img2 * img2, window, stride=1, padding=0, groups=img.shape[1]) - mu2_sq
    )
    sigma12 = (
        F.conv2d(img * img2, window, stride=1, padding=0, groups=img.shape[1]) - mu1_mu2
    )

    cs_map = (2 * sigma12 + c2) / (sigma1_sq + sigma2_sq + c2)
    ssim_map = ((2 * mu1_mu2 + c1) / (mu1_sq + mu2_sq + c1)) * cs_map
    return float(ssim_map.mean([1, 2, 3]))


def _average_precision(
    output: npt.NDArray[np.float64], target: npt.NDArray[np.float64]
) -> float:
    """
    Calculate the average precision of the given inputs.

    See :py:func:`.calculate_map` for details.

    Args:
        output: predicate labels in range :math:`[0, 1]`.
        target: target (inferred) labels in range :math:`[0, 1]`.

    Returns:
        The average precision.
    """
    epsilon = 1e-8

    indices = output.argsort()[::-1]
    total_count = np.cumsum(np.ones((len(output), 1)))

    target_ = target[indices]
    ind = target_ == 1
    pos_count = np.cumsum(ind)
    total = pos_count[-1]
    pos_count[np.logical_not(ind)] = 0
    pp = pos_count / total_count
    precision_at_i = np.sum(pp)
    precision_at_i = precision_at_i / (total + epsilon)

    return precision_at_i


def _mae(pred: npt.NDArray, gt: npt.NDArray, scale: float = 1.0) -> float:
    """
    MAE implementation. See :py:func:`.calculate_mae` for details.

    Args:
        pred: predicate tensor.
        gt: ground-truth tensor.
        scale: the scaling factor used in the tensor data (if any). Defaults to 1.

    Returns:
        The MAE score.
    """
    n = np.prod(gt.shape)
    sum_error = np.sum(np.abs(np.subtract(np.float32(pred), np.float32(gt))))
    mae_error = np.divide(sum_error, n * scale + 1e-4)
    return float(mae_error)


def _mae_torch(pred: torch.Tensor, gt: torch.Tensor, scale: float = 1.0) -> float:
    """
    MAE Torch implementation. See :py:func:`.calculate_mae_torch` for details.

    Args:
        pred: predicate tensor.
        gt: ground-truth tensor.
        scale: the scaling factor used in the tensor data (if any). Defaults to 1.

    Returns:
        The MAE score.
    """
    n = torch.prod(torch.tensor(gt.shape))
    sum_error = torch.sum(torch.absolute(torch.sub(pred.float(), gt.float())))
    mae_error = torch.divide(sum_error, n * scale + 1e-4)
    return float(mae_error)


def calculate_psnr(
    img: npt.NDArray,
    img2: npt.NDArray,
    crop_border: int,
    input_order: str = "HWC",
    test_y_channel: bool = False,
) -> float:
    """
    Calculate PSNR (Peak Signal-to-Noise Ratio).

    Implementation follows: `<https://en.wikipedia.org/wiki/Peak_signal-to-noise_ratio>`__

    Args:
        img: Images with range :math:`[0, 255]`.
        img2: Images with range :math:`[0, 255]`.
        crop_border: Cropped pixels in each edge of an image. These pixels are not
            involved in the calculation.
        input_order: Whether the input order is "HWC" or "CHW". Defaults to "HWC".
        test_y_channel: Test on Y channel of YCbCr. Defaults to false.

    Returns:
        PSNR value.
    """
    assert (
        img.shape == img2.shape
    ), f"Image shapes are different: {img.shape}, {img2.shape}."

    img = functional.convert_to_hwc(img, input_order=input_order)
    img2 = functional.convert_to_hwc(img2, input_order=input_order)

    if crop_border != 0:
        img = img[crop_border:-crop_border, crop_border:-crop_border, ...]
        img2 = img2[crop_border:-crop_border, crop_border:-crop_border, ...]

    if test_y_channel:
        img = functional.to_y_channel(img)
        img2 = functional.to_y_channel(img2)

    img = img.astype(np.float64)
    img2 = img2.astype(np.float64)

    mse = np.mean((img - img2) ** 2)
    if mse == 0:
        return float("inf")
    return 10.0 * np.log10(255.0 * 255.0 / mse)


def calculate_psnr_torch(
    img: torch.Tensor,
    img2: torch.Tensor,
    crop_border: int,
    test_y_channel: bool = False,
) -> float:
    """
    Calculate PSNR (Peak Signal-to-Noise Ratio).

    Implementation follows: `<https://en.wikipedia.org/wiki/Peak_signal-to-noise_ratio>`__

    Args:
        img: Images with range :math:`[0, 255]`.
        img2: Images with range :math:`[0, 255]`.
        crop_border: Cropped pixels in each edge of an image. These pixels are not
            involved in the calculation.
        test_y_channel: Test on Y channel of YCbCr. Defaults to false.

    Returns:
        PSNR value.
    """
    assert (
        img.shape == img2.shape
    ), f"Image shapes are different: {img.shape}, {img2.shape}."

    if crop_border != 0:
        img = img[:, :, crop_border:-crop_border, crop_border:-crop_border]
        img2 = img2[:, :, crop_border:-crop_border, crop_border:-crop_border]

    if test_y_channel:
        img = functional.rgb2ycbcr_torch(img, only_y=True)
        img2 = functional.rgb2ycbcr_torch(img2, only_y=True)

    img = img.to(torch.float64)
    img2 = img2.to(torch.float64)

    mse = torch.mean((img - img2) ** 2, dim=[1, 2, 3])
    return float(10.0 * torch.log10(1.0 / (mse + 1e-8)))


def calculate_ssim(
    img: npt.NDArray,
    img2: npt.NDArray,
    crop_border: int,
    input_order: str = "HWC",
    test_y_channel: bool = False,
) -> float:
    """
    Calculate SSIM (structural similarity).

    Implementation follows: 'Image quality assesment: From error visibility to structural
    similarity'. Results are identical to those of the official MATLAB code in
    `<https://ece.uwaterloo.ca/~z70wang/research/ssim/>`__.
    For three-channel images, SSIM is calculated for each channel and then
    averaged.

    Args:
        img: Images with range :math:`[0, 255]`.
        img2: Images with range :math:`[0, 255]`.
        crop_border: Cropped pixels in each edge of an image. These pixels are not
            involved in the calculation.
        input_order: Whether the input order is "HWC" or "CHW". Defaults to "HWC"
        test_y_channel: Test on Y channel of YCbCr. Defaults to false.

    Returns:
        SSIM.
    """
    assert (
        img.shape == img2.shape
    ), f"Image shapes are different: {img.shape}, {img2.shape}."

    img = functional.convert_to_hwc(img, input_order=input_order)
    img2 = functional.convert_to_hwc(img2, input_order=input_order)

    if crop_border != 0:
        img = img[crop_border:-crop_border, crop_border:-crop_border, ...]
        img2 = img2[crop_border:-crop_border, crop_border:-crop_border, ...]

    if test_y_channel:
        img = functional.to_y_channel(img)
        img2 = functional.to_y_channel(img2)

    img = img.astype(np.float64)
    img2 = img2.astype(np.float64)

    ssims = []
    for i in range(img.shape[2]):
        ssims.append(_ssim(img[..., i], img2[..., i]))
    return np.array(ssims).mean()


def calculate_ssim_torch(
    img: torch.Tensor,
    img2: torch.Tensor,
    crop_border: int,
    test_y_channel: bool = False,
) -> float:
    """
    Calculate SSIM (structural similarity).

    Implementation follows: 'Image quality assesment: From error visibility to structural
    similarity'. Results are identical to those of the official MATLAB code in
    `<https://ece.uwaterloo.ca/~z70wang/research/ssim/>`__.
    For three-channel images, SSIM is calculated for each channel and then
    averaged.

    Args:
        img: Images with range :math:`[0, 255]`.
        img2: Images with range :math:`[0, 255]`.
        crop_border: Cropped pixels in each edge of an image. These pixels are not
            involved in the calculation.
        test_y_channel: Test on Y channel of YCbCr. Defaults to false.

    Returns:
        SSIM.
    """
    assert (
        img.shape == img2.shape
    ), f"Image shapes are different: {img.shape}, {img2.shape}."

    if crop_border != 0:
        img = img[:, :, crop_border:-crop_border, crop_border:-crop_border]
        img2 = img2[:, :, crop_border:-crop_border, crop_border:-crop_border]

    if test_y_channel:
        img = functional.rgb2ycbcr_torch(img, only_y=True)
        img2 = functional.rgb2ycbcr_torch(img2, only_y=True)

    img = img.to(torch.float64)
    img2 = img2.to(torch.float64)

    ssim = _ssim_torch(img * 255.0, img2 * 255.0)
    return ssim


def calculate_mAP(targs: npt.NDArray, preds: npt.NDArray) -> float:
    """
    Calculate the mAP (Mean Average Precision).

    Implementation follows:
    `<https://en.wikipedia.org/wiki/Evaluation_measures_(information_retrieval)#Mean_average_precision>`__

    Args:
        targs: target (inferred) labels in range :math:`[0, 1]`.
        preds: predicate labels in range :math:`[0, 1]`.

    Returns:
        The mAP score
    """
    if np.size(preds) == 0:
        return 0

    class_count = 0
    mAP_sum = 0.0
    for k in range(preds.shape[1]):
        scores = preds[:, k]
        targets = targs[:, k]
        if np.sum(targets) == 0:
            continue
        mAP_sum += _average_precision(scores, targets)
        class_count += 1

    return 100 * (mAP_sum / class_count)


def calculate_mae(pred: npt.NDArray, gt: npt.NDArray, scale: float = 1.0) -> float:
    """
    Compute the MAE (Mean-Average Precision) score.

    Implementation follows: `<https://en.wikipedia.org/wiki/Mean_absolute_error>`__.
    The scale argument is used in the event that the input arrays are not in the range
    :math:`[0, 1]` but instead have been scaled to be in the range :math:`[0, N]` where
    :math:`N` is the factor. For example, if the arrays are images in the range
    :math:`[0, 255]`, then the scaling factor should be set to 255. If the arrays are
    already in the range :math:`[0, 1]`, then the scale can be omitted.

    Args:
        pred: predicate (inferred) array
        gt: ground-truth array
        scale: scaling factor that was used on the input arrays. Defaults to 1.

    Returns:
        The MAE score.
    """
    return _mae(pred, gt, scale)


def calculate_mae_torch(
    pred: torch.Tensor, gt: torch.Tensor, scale: float = 1.0
) -> float:
    """
    Compute the MAE (Mean-Average Precision) score.

    Implementation follows: `<https://en.wikipedia.org/wiki/Mean_absolute_error>`__.
    The scale argument is used in the event that the input arrays are not in the range
    :math:`[0, 1]` but instead have been scaled to be in the range :math:`[0, N]` where
    :math:`N` is the factor. For example, if the arrays are images in the range
    :math:`[0, 255]`, then the scaling factor should be set to 255. If the arrays are
    already in the range :math:`[0, 1]`, then the scale can be omitted.

    Args:
        pred: predicate (inferred) tensor
        gt: ground-truth tensor
        scale: scaling factor that was used on the input tensors. Defaults to 1.

    Returns:
        The MAE score.
    """
    return _mae_torch(pred, gt, scale)



import typing

from torch.optim import lr_scheduler

from helios import core


def _register_default_schedulers(registry: core.Registry) -> None:
    """
    Register the default Torch schedulers to the registry.

    List was obtained from
    https://pytorch.org/docs/stable/optim.html#how-to-adjust-learning-rate

    Args:
        registry: the scheduler registry.
    """
    registry.register(lr_scheduler.LambdaLR)
    registry.register(lr_scheduler.MultiplicativeLR)
    registry.register(lr_scheduler.StepLR)
    registry.register(lr_scheduler.MultiStepLR)
    registry.register(lr_scheduler.ConstantLR)
    registry.register(lr_scheduler.LinearLR)
    registry.register(lr_scheduler.ExponentialLR)
    registry.register(lr_scheduler.PolynomialLR)
    registry.register(lr_scheduler.CosineAnnealingLR)
    registry.register(lr_scheduler.SequentialLR)
    registry.register(lr_scheduler.ReduceLROnPlateau)
    registry.register(lr_scheduler.CyclicLR)
    registry.register(lr_scheduler.OneCycleLR)
    registry.register(lr_scheduler.CosineAnnealingWarmRestarts)


SCHEDULER_REGISTRY = core.Registry("scheduler")
"""
Global instance of the registry for schedulers.

By default, the registry contains the following schedulers:

.. list-table:: Schedulers
    :header-rows: 1

    * - Scheduler
      - Name
    * - ``torch.optim.lr_scheduler.LambdaLR``
      - LambdaLR
    * - ``torch.optim.lr_scheduler.MultiplicativeLR``
      - MultiplicativeLR
    * - ``torch.optim.lr_scheduler.StepLR``
      - StepLR
    * - ``torch.optim.lr_scheduler.MultiStepLR``
      - MultiStepLR
    * - ``torch.optim.lr_scheduler.ConstantLR``
      - ConstantLR
    * - ``torch.optim.lr_scheduler.LinearLR``
      - LinearLR
    * - ``torch.optim.lr_scheduler.ExponentialLR``
      - ExponentialLR
    * - ``torch.optim.lr_scheduler.PolynomialLR``
      - PolynomialLR
    * - ``torch.optim.lr_scheduler.CosineAnnealingLR``
      - CosineAnnealingLR
    * - ``torch.optim.lr_scheduler.SequentialLR``
      - SequentialLR
    * - ``torch.optim.lr_scheduler.ReduceLROnPlateau``
      - ReduceLROnPlateau
    * - ``torch.optim.lr_scheduler.CyclicLR``
      - CyclicLR
    * - ``torch.optim.lr_scheduler.OneCycleLR``
      - OneCycleLR
    * - ``torch.optim.lr_scheduler.CosineAnnealingWarmRestarts``
      - CosineAnnealingWarmRestarts
    * - :py:class:`helios.scheduler.schedulers.CosineAnnealingRestartLR`
      - CosineAnnealingRestartLR
    * - :py:class:`helios.scheduler.schedulers.MultiStepRestartLR`
      - MultiStepRestartLR

Example:
    .. code-block:: python

        import helios.optim as hlo
        import helios.scheduler as hls

        # This automatically registers your optimizer.
        @hls.SCHEDULER_REGISTRY.register
        class MyScheduler:
            ...

        # Alternatively you can manually register a scheduler. like this:
        hls.SCHEDULER_REGISTRY.register(MyScheduler)
"""
_register_default_schedulers(SCHEDULER_REGISTRY)


def create_scheduler(
    type_name: str, *args: typing.Any, **kwargs: typing.Any
) -> lr_scheduler.LRScheduler:
    """
    Create the scheduler for the given type.

    Args:
        type_name: the type of the scheduler to create.
        args: positional arguments to pass into the scheduler.
        kwargs: keyword arguments to pass into the scheduler.

    Returns:
        The scheduler.
    """
    return SCHEDULER_REGISTRY.get(type_name)(*args, **kwargs)



import typing

import torch
from torch import nn
from torch.nn import init
from torch.nn.modules import batchnorm as bn

from helios import core

NETWORK_REGISTRY = core.Registry("network")
"""
Global instance of the registry for networks.

Example:
    .. code-block:: python

        import helios.nn as hln

        # This automatically registers your network.
        @hln.NETWORK_REGISTRY.register
        class MyNetwork:
            ...

        # Alternatively you can manually register a network like this:
        hln.NETWORK_REGISTRY.register(MyNetwork)
"""


def create_network(type_name: str, *args: typing.Any, **kwargs: typing.Any) -> nn.Module:
    """
    Create the network for the given type.

    Args:
        type_name: the type of the network to create.
        args: positional arguments to pass into the network.
        kwargs: keyword arguments to pass into the network.
    Returns:
        The network.
    """
    return NETWORK_REGISTRY.get(type_name)(*args, **kwargs)


@torch.no_grad()
def default_init_weights(
    module_list: list[nn.Module] | nn.Module,
    scale: float = 1,
    bias_fill: float = 0,
    **kwargs: typing.Any,
) -> None:
    """
    Initialize network weights.

    Specifically, this function will default initialize the following types of blocks:

    * ``torch.nn.Conv2d``,
    * ``torch.nn.Linear``,
    * ``torch.nn.modules.batchnorm._BatchNorm``

    Args:
        module_list: the list of modules to initialize.
        scale: scale initialized weights, especially for residual blocks. Defaults to 1.
        bias_fill: bias fill value. Defaults to 0.
        kwargs: keyword arguments for the ``torch.nn.init.kaiming_normal_`` function.
    """
    if not isinstance(module_list, list):
        module_list = [module_list]

    for module in module_list:
        for m in module.modules():
            if isinstance(m, nn.Conv2d | nn.Linear):
                init.kaiming_normal_(m.weight, **kwargs)
                m.weight.data *= scale
                if m.bias is not None:
                    m.bias.data.fill_(bias_fill)
            elif isinstance(m, bn._BatchNorm):  # noqa: SLF001
                init.constant_(m.weight, 1)
                if m.bias is not None:
                    m.bias.data.fill_(bias_fill)



try:
    import optuna
except ImportError as e:
    raise ImportError("error: OptunaPlugin requires Optuna to be installed") from e
import gc
import pathlib
import pickle
import shutil
import typing

import torch

import helios.core.distributed as dist
import helios.model as hlm
import helios.plugins as hlp
import helios.trainer as hlt

# Ignore private member access.
# ruff: noqa: SLF001

_PRUNED_KEY = "ddp_hl:pruned"
_CYCLE_KEY = "ddp_hl:cycle"
_ORIG_NUMBER_KEY = "hl:id"


def _backup_study(storage_path: pathlib.Path) -> None:
    if not storage_path.exists():
        return

    name = storage_path.stem
    root = storage_path.parent

    i = -1
    for path in root.iterdir():
        if not path.is_file():
            continue
        if "_bkp-" in path.stem:
            elems = path.stem.split("_")
            idx = [i for i, e in enumerate(elems) if "bkp-" in e][0]
            n = int(elems[idx].split("-")[-1])
            i = max(i, n)

    i = i + 1
    bkp_path = root / f"{name}_bkp-{i}.db"
    shutil.copy(str(storage_path), str(bkp_path))


def resume_study(
    study_args: dict[str, typing.Any],
    failed_states: typing.Sequence = (optuna.trial.TrialState.FAIL,),
    backup_study: bool = True,
) -> optuna.Study:
    """
    Resume a study that stopped because of a failure.

    The goal of this function is to allow studies that failed due to an error (either an
    exception, system error etc.) to continue utilising the built-in checkpoint system
    from Helios. To accomplish this, the function will do the following:

    #. Grab all the trials from the study created by the given ``study_args``, splitting
        them into three groups: completed/pruned, failed, and failed but completed.
    #. Create a new study with the same name and storage. This new study will get all of
        the completed trials of the original, and will have the failed trials re-enqueued.

    .. warning::
        This function **requires** the following conditions to be true:
            #. It is called **before** the trials are started.
            #. The study uses ``RDBStorage`` as the storage argument for
            ``optuna.create_study``.
            #. ``load_if_exists`` is set to True in ``study_args``.
            #. ``TrialState.PRUNED`` **cannot** be in the list of ``failed_states``.

    The ``failed_states`` argument can be used to set additional trial states to be
    considered as "failures". This can be useful when dealing with special cases where
    trials were either completed or pruned but need to be re-run.

    By default, the original study (assuming there is one) will be backed up with the name
    ``<study-name>_backup-#`` where ``<study-name>`` is the name of the database of the
    original study, and ``#`` is an incremental number starting at 0. This behaviour can
    be disabled by setting ``backup_study`` to False.

    This function works in tandem with
    :py:meth:`~helios.plugins.optuna.OptunaPlugin.configure_model` to ensure that when
    the failed trial is re-run, the original save name is restored so any saved
    checkpoints can be re-used so the trial can continue instead of starting from
    scratch.

    .. note::
        Only trials that fail but haven't been completed will be enqueued by this
        function. If a trial fails and is completed later on, it will be treated as if it
        had finished successfully.

    Args:
        study_args: dictionary of arguments for ``optuna.create_study``.
        failed_states: the trial states that are considered to be failures and should
            be re-enqueued.
        backup_study: if True, the original study is backed up so it can be re-used later
            on.
    """
    if "storage" not in study_args:
        raise TypeError("error: RDB storage is required for resuming studies")
    if "load_if_exists" not in study_args or not study_args["load_if_exists"]:
        raise KeyError("error: study must be created with 'load_if_exists' set to True")
    if optuna.trial.TrialState.PRUNED in failed_states:
        raise ValueError("error: pruned trials cannot be considered as failed")

    storage_str: str = study_args["storage"]
    if not isinstance(storage_str, str):
        raise TypeError("error: only strings are supported for 'storage'")

    storage = pathlib.Path(storage_str.removeprefix("sqlite:///")).resolve()
    if backup_study:
        _backup_study(storage)

    # Step 1: create the study with the current DB and grab all the trials.
    study = optuna.create_study(**study_args)

    # Fast exit: if there are no trials, return immediately.
    if len(study.trials) == 0:
        return study

    complete: list[optuna.trial.FrozenTrial] = []
    failed_but_completed: list[optuna.trial.FrozenTrial] = []
    failed: dict[int, optuna.trial.FrozenTrial] = {}

    for trial in study.trials:
        if (
            trial.state == optuna.trial.TrialState.COMPLETE
            and _ORIG_NUMBER_KEY in trial.user_attrs
        ):
            failed_but_completed.append(trial)
        elif (
            trial.state == optuna.trial.TrialState.COMPLETE
            or trial.state == optuna.trial.TrialState.PRUNED
        ):
            complete.append(trial)
        elif trial.state in failed_states:
            failed[trial.number] = trial

    # Make sure that any trials that failed but were completed are pruned from the failed
    # trials list.
    for trial in failed_but_completed:
        trial_num = trial.user_attrs[_ORIG_NUMBER_KEY]
        failed.pop(trial_num, None)

    # Make sure that the study is cleared out before we attempt to rename the storage
    del study
    gc.collect()

    # Step 2: rename the DB and create a new empty study.
    tmp_storage = storage.parent / (storage.stem + "_tmp" + storage.suffix)
    storage.rename(tmp_storage)
    study = optuna.create_study(**study_args)

    # Step 3: move all the trials into the new study, re-setting all trials that failed.
    for trial in complete:
        study.add_trial(trial)

    for _, trial in failed.items():
        study.enqueue_trial(trial.params, {_ORIG_NUMBER_KEY: trial.number})

    # Once everything's done, clean up the temp storage.
    tmp_storage.unlink()

    return study


def checkpoint_sampler(trial: optuna.Trial, chkpt_root: pathlib.Path) -> None:
    """
    Create a checkpoint with the state of the sampler.

    This function can be used to ensure that if a study is restarted, the state of the
    sampler is recovered so trials can be reproducible. The function will automatically
    create a checkpoint using ``torch.save``.

    .. note::
        It is recommended that this function be called at the start of the objective
        function to ensure the checkpoint is made correctly, but it can be called at any
        time.

    Args:
        trial: the current trial.
        chkpt_root: the root where the checkpoints will be saved.
    """
    chkpt_path = chkpt_root / (f"sampler_trial-{trial.number}.pkl")

    sampler = trial.study.sampler
    with chkpt_path.open("wb") as outfile:
        pickle.dump(sampler, outfile)


def restore_sampler(chkpt_root: pathlib.Path) -> optuna.samplers.BaseSampler | None:
    """
    Restore the sampler from a previously saved checkpoint.

    This function can be used in tandem with
    :py:func:`~helios.plugins.optuna.checkpoint_sampler` to ensure that the last
    checkpoint is loaded and the correct state is restored for the sampler. This function
    **needs** to be called before ``optuna.create_study`` is called.

    Args:
        chkpt_root: the root where the checkpoints are stored.

    Returns:
        The restored sampler.
    """

    def key(path: pathlib.Path) -> int:
        return int(path.stem.split("-")[-1])

    chkpts = list(chkpt_root.glob("*.pkl"))
    chkpts.sort(key=key)
    if len(chkpts) == 0:
        return None

    sampler = pickle.load(chkpts[-1].open("rb"))
    return sampler


@hlp.PLUGIN_REGISTRY.register
class OptunaPlugin(hlp.Plugin):
    """
    Plug-in to do hyper-parameter tuning with Optuna.

    This plug-in integrates `Optuna <https://optuna.readthedocs.io/en/stable/>`__ into the
    training system in order to provide hyper-parameter tuning. The plug-in provides the
    following functionality:

    #. Automatic handling of trial pruning.
    #. Automatic reporting of metrics.
    #. Exception registration for trial pruning.
    #. Easy integration with Helios' checkpoint system to continue stopped trials.

    .. warning::
        This plug-in **requires** Optuna to be installed before being used. If it isn't,
        then :py:exc:`ImportError` is raised.

    Example:
        .. code-block:: python

            import helios.plugins as hlp
            import optuna

            def objective(trial: optuna.Trial) -> float:
                datamodule = ...
                model = ...
                plugin = hlp.optuna.OptunaPlugin(trial, "accuracy")

                trainer = ...

                # Automatically registers the plug-in with the trainer.
                plugin.configure_trainer(trainer)

                # This can be skipped if you don't want the auto-resume functionality or
                # if you wish to manage it yourself.
                plugin.configure_model(model)

                trainer.fit(model, datamodule)
                plugin.check_pruned()
                return model.metrics["accuracy"]

            def main():
                # Note that the plug-in requires the storage to be persistent.
                study = optuna.create_study(storage="sqlite:///example.db", ...)
                study.optimize(objective, ...)

    Args:
        trial: the Optuna trial.
        metric_name: the name of the metric to monitor. This assumes the name will be
            present in the :py:attr:`~helios.model.model.Model.metrics` table.
    """

    plugin_id = "optuna"

    def __init__(self, trial: optuna.Trial, metric_name: str) -> None:
        """Create the plug-in."""
        super().__init__(self.plugin_id)
        self._trial = trial
        self._metric_name = metric_name
        self._last_cycle: int = 0

        self.unique_overrides.should_training_stop = True

    @property
    def trial(self) -> optuna.Trial:
        """Return the trial."""
        return self._trial

    @trial.setter
    def trial(self, t: optuna.Trial) -> None:
        self._trial = t

    def configure_trainer(self, trainer: hlt.Trainer) -> None:
        """
        Configure the trainer with the required settings.

        This will do two things:

        #. Register the plug-in itself with the trainer.
        #. Append the trial pruned exception to the trainer.

        Args:
            trainer: the trainer instance.
        """
        self._register_in_trainer(trainer)
        self._append_train_exceptions(optuna.TrialPruned, trainer)

    def configure_model(self, model: hlm.Model) -> None:
        """
        Configure the model to set the trial number into the save name.

        This will alter the :py:attr:`~helios.model.model.Model.save_name` property of the
        model by appending :code:`_trial-<trial-numer>`.

        Args:
            model: the model instance.
        """
        n_trial = self.trial.user_attrs.get(_ORIG_NUMBER_KEY, self.trial.number)
        model._save_name = model._save_name + f"_trial-{n_trial}"

    def suggest(self, type_name: str, name: str, **kwargs: typing.Any) -> typing.Any:
        """
        Generically Wrap the ``suggest_`` family of functions of the optuna trial.

        This function can be used to easily invoke the corresponding ``suggest_`` function
        from the Optuna trial held by the plug-in without having to manually type each
        individual function. This lets you write generic code that can be controlled by an
        external source (such as command line arguments or a config table). The function
        wraps the following functions:

        .. list-table:: Suggestion Functions
            :header-rows: 1

            * - Function
              - Name
            * - ``optuna.Trial.suggest_categorical``
              - categorical
            * - ``optuna.Trial.suggest_int``
              - int
            * - ``optuna.Trial.suggest_float``
              - float

        .. warning::
            Functions that are marked as deprecated by Optuna are *not* included in this
            wrapper.

        .. note::
            You can find the exact arguments for each function `here
            <https://optuna.readthedocs.io/en/stable/reference/generated/optuna.trial.Trial.html>`__.

        Example:
            .. code-block:: python

                import helios.plugin as hlp
                import optuna

                def objective(trial: optuna.Trial) -> float:
                    plugin = hlp.optuna.OptunaPlugin(trial, "accuracy")
                    # ... configure model and trainer.

                    val1 = plugin.suggest("categorical", "val1", choices=[1, 2, 3])
                    val2 = plugin.suggest("int", "val2", low=0, high=10)
                    val3 = plugin.suggest("float", "val3", low=0, high=1)

        Args:
            type_name: the name of the type to suggest from.
            name: a parameter name
            **kwargs: keyword arguments to the corresponding suggest function.

        Raises:
            KeyError: if the value passed in to ``type_name`` is not recognised.
        """
        if type_name not in ("categorical", "float", "int"):
            raise KeyError(f"error: {type_name} is not a valid suggestion type.")

        fn = getattr(self._trial, f"suggest_{type_name}")
        return fn(name, **kwargs)

    def setup(self) -> None:
        """
        Configure the plug-in.

        Raises:
            ValueError: if the study wasn't created with persistent storage.
        """
        if self.is_distributed and not (
            isinstance(self.trial.study._storage, optuna.storages._CachedStorage)
            and isinstance(self.trial.study._storage._backend, optuna.storages.RDBStorage)
        ):
            raise ValueError(
                "error: optuna integration supports only optuna.storages.RDBStorage "
                "in distributed mode"
            )

    def report_metrics(self, validation_cycle: int) -> None:
        """
        Report metrics to the trial.

        This function should be called from the model once the corresponding metrics have
        been saved into the :py:attr:`~helios.model.model.Model.metrics` table.

        Example:
            .. code-block:: python

                import helios.model as hlm
                import helios.plugins.optuna as hlpo

                class MyModel(hlm.Model):
                    ...
                    def on_validation_end(self, validation_cycle: int) -> None:
                        # Compute metrics
                        self.metrics["accuracy"] = 10

                        plugin = self.trainer.plugins[hlpo.OptunaPlugin.plugin_id]
                        assert isinstance(plugin hlpo.OptunaPlugin)
                        plugin.report_metrics(validation_cycle)

        .. note::
            In distributed training, only rank 0 will report the metrics to the trial.

        Args:
            validation_cycle: the current validation cycle.
        """
        model = self.trainer.model
        if not model.metrics or self._metric_name not in model.metrics:
            return

        if self.rank == 0:
            self.trial.report(model.metrics[self._metric_name], validation_cycle)
        self._last_cycle = validation_cycle

    def should_training_stop(self) -> bool:
        """
        Handle trial pruning.

        Returns:
            True if the trial should be pruned, false otherwise.
        """
        should_stop = False
        if self.rank == 0:
            should_stop = self.trial.should_prune()

        # Sync the value across all processes (if using distributed training).
        if self.is_distributed:
            t = dist.all_reduce_tensors(torch.tensor(should_stop).to(self.device))
            should_stop = t.item()  # type: ignore[assignment]

        if should_stop and self.rank == 0:
            self.trial.set_user_attr(_PRUNED_KEY, True)
            self.trial.set_user_attr(_CYCLE_KEY, self._last_cycle)

        return should_stop

    def on_training_end(self) -> None:
        """
        Clean-up on training end.

        If training is non-distributed and the trial was pruned, then this function will
        do the following:

        #. Call :py:meth:`~helios.model.model.Model.on_training_end` to ensure metrics are
           correctly logged (if using).
        #. Raise :py:exc:`optuna.TrialPruned` exception to signal the trial was pruned.

        If training is distributed, this function does nothing.

        Raises:
            TrialPruned: if the trial was pruned.
        """
        if not self.is_distributed and self.trial.should_prune():
            self.trainer.model.on_training_end()
            raise optuna.TrialPruned(f"Pruned on validation cycle {self._last_cycle}")

    def check_pruned(self) -> None:
        """
        Ensure pruned distributed trials are correctly handled.

        Due to the way distributed training works, we can't raise an exception within the
        distributed processes, so we have to do it after we return to the main process.
        If the trial was pruned, this function will raise :py:exc:`optuna.TrialPruned`. If
        distributed training wasn't used, this function does nothing.

        .. warning::
            You *must* ensure this function is called after
            :py:meth:`~helios.trainer.Trainer.fit` to ensure pruning works correctly.

        Raises:
            TrialPruned: if the trial was pruned.
        """
        trial_id = self.trial._trial_id
        study = self.trial.study
        trial = study._storage._backend.get_trial(trial_id)  # type: ignore[attr-defined]
        is_pruned = trial.user_attrs.get(_PRUNED_KEY)
        val_cycle = trial.user_attrs.get(_CYCLE_KEY)

        if is_pruned is None or val_cycle is None:
            return

        if is_pruned:
            raise optuna.TrialPruned(f"Pruned on validation cycle {val_cycle}")

    def state_dict(self) -> dict[str, typing.Any]:
        """
        Get the state of the current trial.

        This will return the parameters to be optimised for the current trial.

        Returns:
            The parameters of the trial.
        """
        return self._trial.params
